from __future__ import absolute_import

import os
import sys

# If we are running from a wheel, add the wheel to sys.path
# This allows the usage python pip-*.whl/pip install pip-*.whl
if __package__ == '':
    # __file__ is pip-*.whl/pip/__main__.py
    # first dirname call strips of '/__main__.py', second strips off '/pip'
    # Resulting path is the name of the wheel itself
    # Add that to sys.path so we can import pip9
    path = os.path.dirname(os.path.dirname(__file__))
    sys.path.insert(0, path)

import pip9  # noqa

if __name__ == '__main__':
    sys.exit(pip9.main())
